#  A mud client for
#  Dawn of the Ages  http://dawn.rplay.net
#
#  Copyright (C) 2001,2002 Peter Herth <p.herth@uni-koeln.de>
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from Tkinter import *
from tkMessageBox import *
from tkSimpleDialog import *
from tkFileDialog import *
from Utilities import *

import tkFont
import tkColorChooser

import re
import string
import os

import Term
import Client
import copy
def make_entry(master,text="",column=0,row=0):
    """Creates an Entry-field with a label """
    Label(master,text=text,justify=LEFT).grid(column=2*column,row=2*row,sticky=W)
    entry = Entry(master)
    entry.grid(column=2*column,row=2*row+1,sticky=E+W)
    return entry

def settext(entry,text):
    """replaces the text of an entry-field by new text """
    entry.delete(0,END)
    entry.insert(END,text)
   
class FontSelector(Toplevel):
    def __init__(self,  gui):
        Toplevel.__init__(self)
        self.title("Select font")
        self.gui = gui
        self.transient(gui)

        f = Frame (self)
        f1 = Frame (f)
        self.bold = 0
        self.font = self.gui.fontname
        self.fontname = self.font[0]
        self.fontsize = self.font[1]
        if len(self.font)>2:
            if self.font[2]=="bold":
                self.bold = 1

        list = Listbox(f1)
        list.pack(side=LEFT,fill=BOTH,expand=YES)
        self.list = list
        self.scrollbar = Scrollbar(f1, orient=VERTICAL)
        self.scrollbar.config(command=self.list.yview)
        self.scrollbar.pack(side=LEFT,fill=Y)
        list.configure(yscrollcommand=self.scrollbar.set)
        #bind button-release instead of button-press to catch drags
        self.list.bind("<ButtonRelease-1>", self.list_has_changed)
        #f.pack(side=TOP, fill=BOTH,expand=YES)
        f1.pack(side=LEFT,expand=YES, fill=BOTH)
        #f.grid(row=0,column=0,sticky=W)
        #self.columnconfigure(0, weight=1)
        #self.rowconfigure(0, weight=1)
        #self.columnconfigure(1, weight=1)

        f1= Frame(f)

        self.scale = Scale(f1, orient=VERTICAL, length=200, from_=2, to = 48,command=self.size_change)
        self.scale.pack(side=TOP,fill=Y,expand=YES)
        #self.scale.grid(row=0,column=1)

        self.boldvar=IntVar()
        


        self.boldbox = Checkbutton(f1, text="Bold", var=self.boldvar, command=self.bold_change)
        if self.bold:
            self.boldvar.set(1)
        self.boldbox.pack(side=TOP)

        self.scale.set(self.fontsize)
        f1.pack(side=LEFT,fill=Y)
        #f.grid(row=0, column=1)
        f.pack(side = TOP,expand=YES,fill=BOTH)
        f= Frame(self)

        self.demofield=Text(f, width = 25, height=8)
        self.demofield.tag_config("f", font=self.gui.fontname)
        self.demofield.insert(END,"""ABCDE abcde XYZ xyz\naaaa           z\naiii           z\nammm           z""","f")
        #self.demofield.grid(row=1,column=0)
        self.demofield.pack(fill=BOTH,expand=YES)
        f.pack(side=TOP,fill=X)
        #f.grid(row=1,column=0, columnspan=2)



        
        f = Frame(self)
        b = Button(f, text="OK", width=10,command= self.ok, default=ACTIVE)
        b.pack(side=LEFT)
        b = Button(f, text="Apply", width=10,command= self.apply)
        b.pack(side=LEFT)
        b = Button(f, text="Cancel", width=10,command= self.cancel)
        b.pack(side=LEFT)
        self.bind("<Escape>", self.cancel)
        
        f.pack(side=TOP, fill=X)
        #f.grid(row=2,column=0, columnspan=2,sticky=W)


        fams=[]
        for f in tkFont.families():
            fams.append(f)
        fams.sort()
        i = 0
        pos = 0
        for f in fams:
            i = i + 1
            list.insert(END,f)
            if f==self.fontname:
                pos = i

        if pos:
            list.selection_set(pos-1)
            list.see(pos-1)

    def changefont(self):
        if self.bold:            
            self.font = (self.fontname, self.fontsize, "bold")
        else:
            self.font = (self.fontname, self.fontsize, "normal")
        #print "fontname=",self.font
        
        self.demofield.tag_config("f",font=self.font)

    def bold_change(self):
        self.bold = self.boldvar.get()
        self.changefont()
        #print "bold:", bold

    def size_change(self, event):
        self.fontsize = self.scale.get()
        self.changefont()
        
    def list_has_changed(self,event):
        selected = self.list.curselection()
        #print "selected=",selected
        if len(selected)>0:
            index = int(selected[0])
            
            #print "index=",index
            
            self.fontname = self.list.get(index)
            self.changefont()
        
    def ok(self):
        self.gui.setfont(self.font)
        self.gui.focus_set()
        self.destroy()
        
    def apply(self):
        self.gui.setfont(self.font)
        
    
    def cancel(self):
        self.gui.focus_set()
        self.destroy()
    



class AddToDb(Toplevel):
    def __init__(self,  gui, client):
        Toplevel.__init__(self)
        self.title("Add Character")
        self.gui = gui
        self.transient(gui)
        self.client = client
        
        Label(self,text="display name:").grid(row=0,column=0,sticky=W)
        Label(self,text="player name:").grid(row=1,column=0,sticky=W)
        Label(self,text="password:").grid(row=2,column=0,sticky=W)
        Label(self,text="retype password:").grid(row=3,column=0,sticky=W)        

        self.dname=Entry(self)
        self.pname=Entry(self)
        self.pw=Entry(self,show="*")
        self.pwc=Entry(self,show="*")
        self.dname.grid(row=0,column=1,sticky=EW)
        self.pname.grid(row=1,column=1,sticky=EW)
        self.pw.grid(row=2,column=1,sticky=EW)
        self.pwc.grid(row=3,column=1,sticky=EW)

        self.columnconfigure(1,weight=1)
        f = Frame(self)
        Button(f, text="Ok",command=self.ok).pack(side=LEFT)
        Button(f, text="Cancel",command=self.cancel).pack(side=LEFT)

        f.grid(row=4,column=1)
        self.dname.focus_set()

    def ok(self):
        dname = self.dname.get()
        pname = self.pname.get()
        pw = self.pw.get()
        pwc = self.pwc.get()

        if not dname or len(dname.strip())<1:
            showinfo("Add Character", "You need to provide a name to be displayed.")
            return
        if not pname or len(pname.strip())<1:
            showinfo("Add Character", "You need to provide a player name.")
            return
        if not pw or len(pw.strip())<1:
            showinfo("Add Character", "You need to provide a password.")
            return

        if not pwc or len(pwc.strip())<1:
            showinfo("Add Character", "You need to repeat the password.")
            return

        pw = pw.strip()
        pwc = pwc.strip()
        if pw != pwc:
            showinfo("Add Character", "The passwords don't match.")
            return

        self.client.add_to_chardb(dname.strip(), pname.strip(), pw)
        self.gui.add_char(dname.strip())
        self.withdraw()
        pass

    def cancel(self):
        self.withdraw()
    
class Dialog(Toplevel):
    def __init__(self, parent,title = None):
        Toplevel.__init__(self, parent)
        self.transient(parent)
        # window doesnt appear in the taskbar and is iconified with
        # the parent window
        if title:
            self.title(title)

        self.parent = parent
        self.result = None

        #call the method to create the body-frame, which can be overwritten in sub-classes
        body_frame = self.create_body(self)
        self.initial_focus =body_frame
        body_frame.grid(sticky=E+W+N+S,padx=5,pady=5)

        #call the method to create the buttonbox, which can be overwritten in sub-classes
        self.buttonbox()
        #self.grab_set() #makes the dialog modal

        if not self.initial_focus:
            self.initial_focus = self
        self.protocol("WM_DELETE_WINDOW", self.cancel) #make closing=cancel
        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                  parent.winfo_rooty()+50))

        self.initial_focus.focus_set()
        #self.wait_window(self)
    #
    # construction hooks

    def create_body(self, master):
        # create dialog body.  return widget that should have
        # initial focus.  this method should be overridden
        return Frame(master)


    def buttonbox(self):
        # add standard button box. override if you don't want the
        # standard buttons
        
        box = Frame(self)

        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.grid(row=0,column=0, padx=5, pady=5)
        w = Button(box, text="Cancel", width=10, command=self.cancel)
        w.grid(row=0,column=1, padx=5, pady=5)
        self.bind("<Escape>", self.cancel)
        box.grid(sticky=W+E)
        return box
    #
    # standard button semantics

    def ok(self, event=None):

        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return
            
        self.withdraw()
        self.update_idletasks()
        self.apply()
        self.parent.focus_set()
        self.destroy()

    def cancel(self, event=None):
        # put focus back to the parent window

        self.parent.focus_set()
        self.destroy()

    #
    # command hooks

    def validate(self):

        return 1 # override

    def apply(self):
        print client.triggers
        pass # override



class HashFrame(Frame):
    # a Frame to display and edit a hash which contains a key with associated commands (triggers, speedwalks, aliases)
    #arguments:
    #master: parent frame
    #hashlist: the hash with the keys and associated commands
    #name: What an item of the list is called (eg. "trigger", "command"), used to create the "add" dialog
    
  def __init__(self, master,hashlist,name):
        Frame.__init__(self,master)
        self.current=None
        self.hash=hashlist
        self.name=name
        self.scrollbar = Scrollbar(self, orient=VERTICAL)
        self.list=Listbox(self,exportselection=0,yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.list.yview)
        self.list.grid(row=0,column=0,sticky=S+N)
        self.scrollbar.grid(row=0,column=1,sticky=S+N)
	self.list.bind("<Button-1>", self.list_has_changed)
        #the list widget has some wierd standard-bindings for these events
        self.list.bind("<B1-Motion>",self.list_ignore)
        self.list.bind("<Double-Button-1>",self.list_ignore)
        #bind button-release to mark the right entry after a drag
        self.list.bind("<ButtonRelease-1>",self.list_ignore)
        self.text=Text(self)
        self.text.grid(row=0,column=2)
        self.text.bind("<FocusOut>",self.make_hash) #if focus is lost, update the associated hash
        
        keys = self.hash.keys()
        keys.sort()
        for k in keys:
            self.list.insert(END,k)
        self.list.activate(0)
        self.list.select_clear(0,END)
        self.list.select_set(0)
        self.make_text()
        self.make_buttons(name)

  def re_init(self,hashlist):
      #to re-initialize the frame with another hash
      self.text.delete(1.0,END)
      self.list.delete(0,END)
      self.hash=hashlist
      keys = self.hash.keys()
      keys.sort()
      for k in keys:
          self.list.insert(END,k)
      self.list.select_set(0)
      self.make_text()
      
  def make_buttons(self,name):
        #Buttons in this frame
        buttonframe=Frame(self)
        buttonframe.grid(row=1,column=0,columnspan=3)
        self.add_button=Button(buttonframe,text="Add "+name,command=self.add_hash)
        self.del_button=Button(buttonframe,text="Delete "+name,command=self.del_hash)
        self.cp_button=Button(buttonframe,text="Copy "+name,command=self.cp_hash)
        self.add_button.grid(row=0,column=0,padx=20,sticky=E+W)
        self.del_button.grid(row=0,column=1)
        self.cp_button.grid(row=0,column=2)
        

  def add_hash(self):
      """Adds a new item to the hash
      """
      self.make_hash()
      newname=askstring("New "+self.name,"Name of new "+self.name,parent=self)
      if self.hash.has_key(newname):
          if not askyesno("New "+self.name, "Overwrite old "+self.name+" "+newname+"?"):
              return
      #add the item
      self.hash[newname]=""
      #empty the list
      self.list.delete(0, END)
      #and insert the keys anew in the right order
      keys = self.hash.keys()
      self.text.delete(1.0, END)
      keys.sort()
      i = 0
      for k in keys:
          self.list.insert(END,k)
          if (k == newname):
              idx = i
          i = i + 1
      #move selection to newly-created item
      self.list.activate(idx)
      self.list.select_set(idx)
      #and set the text-box accordingly
      self.make_text()
           
  def del_hash(self):
      """Removes the currently selected item from the hash
      """
      self.make_hash()
      idx=self.list.index(ACTIVE)
      name=self.list.get(ACTIVE)
      del self.hash[name]
      #remove all items from the list
      self.list.delete(0, END)
      #and insert them anew
      keys = self.hash.keys()
      keys.sort()
      for k in keys:
          self.list.insert(END,k)

      #move selection to the item behind the deleted one
      self.list.activate(idx)
      self.list.select_clear(0,END)
      self.list.select_set(idx)
      self.list.see(idx)

      #and set the text-box accordingly

      self.text.delete(1.0, END)
      self.make_text()
      
  def cp_hash(self):
      """Copies the selected item to a new one
      """
      self.make_hash()
      idx=self.list.index(ACTIVE)
      newname=askstring("Copy "+self.name,"Name of new "+self.name,parent=self)
      if self.hash.has_key(newname):
          if not askyesno("Copy "+self.name, "Overwrite present "+self.name+" "+newname+"?"):
              return
      self.hash[newname]=self.hash[self.list.get(idx)]
      self.list.delete(0, END)
      keys = self.hash.keys()
      keys.sort()
      i=0
      for k in keys:
          self.list.insert(END,k)
          if (k == newname):
              idx = i
          i = i + 1
      self.list.activate(idx)
      self.list.select_clear(0,END)
      self.list.select_set(idx)
      self.list.see(idx)
      self.text.delete(1.0, END)
      self.make_text()

  
  def list_has_changed(self, event):
      #parse the commands in the text-field
      self.make_hash()
      #clear the text-field
      self.text.delete(1.0, END)
      #activate the appropriate item in the list
      self.list.activate(self.list.nearest(event.y))
      #and fill the text-field
      self.make_text()
      
  def make_hash(self,event=None):
      #replaces newlines with semicolons, removes double newlines, removes trailing semicolons
      index = self.list.get(ACTIVE)
      if index:
          self.hash[index]=re.sub("(^ *)|( *$)","",string.replace(string.replace(self.text.get(1.0,END),"\n\n","\n"),"\n",";"))
          if self.hash[index][-1] == ";":
              self.hash[index]=self.hash[index][:-1]
          
          
  def list_ignore(self, event):
      return "break"
  
  def make_text(self):
      """replaces non-escaped semi-colons with newlines for nicer output
      """
      index = self.list.get(ACTIVE)
      if index:
         self.text.insert(END,re.sub('([^\\\]);',"\\1\\n",self.hash[self.list.get(ACTIVE)]))
         


#A dialog-class to edit a hash containing a key and associated commands, uses the HashFrame class
class HashDialog(Dialog):
    
    def __init__(self, parent, Hash,name,title):
        #make a backup copy of the hash
        self.backuphash=Hash.copy()
        self.hash=Hash
        self.name=name
        Dialog.__init__(self,parent,title)
       
    def create_body(self, master):
        self.aliasframe = HashFrame(master,self.hash,self.name)
        self.aliasframe.pack()
        return self.aliasframe
    
    def apply(self):
        #index = self.aliasframe.list.get(ACTIVE)
        #apply the current edits in the text-entry
        #self.hash[index]=re.sub("(^ *)|( *$)","",string.replace(string.replace(self.aliasframe.text.get(1.0,END),"\n\n","\n"),"\n",";"))
        #if self.hash[index][-1] == ";":
        #    self.hash[index]=self.hash[index][:-1]
        
        #self.hash.clear()
        #for k in self.localhash.keys():
        #    self.hash[k] = self.localhash[k]
        self.gui.client.output(""+self.hash)
        return None

    def re_init(self):
        #restore from backup
        self.hash.clear()
        for k in self.backuphash.keys():
            self.hash[k] = self.backuphash[k]
        self.aliasframe.re_init(self.hash)

    def cancel(self):
        #restore hash from backup and close dialog
        self.hash.clear()
        for k in self.backuphash.keys():
            self.hash[k] = self.backuphash[k]
        Dialog.cancel(self)
            
       
## class ButtonsDialog(HashDialog):
##     def __init__(self,parent,client, name,title):
##         self.client = client
##         self.backuphash={}
##         for k in client.buttons.keys():
##             self.backuphash[k]=(buttons[k].commandstring)
##         self.hash=client.buttons
##         self.name=name
##         Dialog.__init__(self,parent,title)
        
##     def buttonbox(self):
##         box=HashDialog.buttonbox(self)
##         Button(box, text="Wizard...", width=10, command=self.button_wizard).grid(row=0,column=2, padx=5, pady=5)
##         return box
    
##     def apply(self):
##         index = self.aliasframe.list.get(ACTIVE)
##         self.hash[index]=re.sub("(^ *)|( *$)","",string.replace(string.replace(self.aliasframe.text.get(1.0,END),"\n\n","\n"),"\n",";"))
##         for k in self.client.buttons.keys():
##             remove_button(k)
##         for k in self.hash.keys():
##             make_button(k+"~"+self.hash[k])
##     def button_wizard(self):
##         ButtonWizard(self)


##     def re_init(self):
##         self.localhash={}
##         for k in self.hash.keys():
##             self.localhash[k]=(self.client.buttons[k].commandstring)
##         self.aliasframe.re_init(self.localhash)

##     def cancel(self):
##         print "Canceling"
##         for k in self.client.buttons.keys():
##             remove_button(k)
##         for k in self.backuphash.keys():
##             make_button(k,"~"+self.backuphash[k])
##         HashDialog.cancel(self) 
        
## class ButtonWizard(Dialog):

##     def body(self, master,triggers):

##         self.spellscrollbar = Scrollbar(master, orient=VERTICAL)
##         self.spellscrollbar.grid(row=1,column=1,sticky=S+N,rowspan=4)
##         self.spelllist=Listbox(master,exportselection=0,yscrollcommand=self.spellscrollbar.set,selectmode=MULTIPLE)
##         self.spelllist.grid(row=1,column=0,sticky=S+N,rowspan=4)
##         self.spellscrollbar.config(command=self.spelllist.yview)
##         self.selscrollbar = Scrollbar(master, orient=VERTICAL)
##         self.selscrollbar.grid(row=1,column=4,sticky=S+N,rowspan=4)
##         self.sellist=Listbox(master,exportselection=0,yscrollcommand=self.selscrollbar.set,selectmode=MULTIPLE)
##         self.sellist.grid(row=1,column=3,sticky=S+N,rowspan=4)
##         self.selscrollbar.config(command=self.sellist.yview)
##         Button(master,text=">>",command=self.select).grid(row=2,column=2)
##         Button(master,text="<<",command=self.deselect).grid(row=3,column=2)
##         Label(master,text="Affects").grid(row=0,column=0)
##         Label(master,text="Selected").grid(row=0,column=3)
##         Label(master,text="Attention!\nThis wizard adds not only buttons,\nbut also triggers to activate and deactivate them.\nBe sure to save the triggers, too.",justify=CENTER).grid(row=5,column=0,columnspan=4,sticky=E+W)
##         v = affects.keys()
##         v.sort()
##         for k in v:
##             self.spelllist.insert(END,"x"+k)
        
##     def select(self):
##         items = self.spelllist.curselection()
##         try:
##             items = map(int, items)
##         except ValueError: pass
##         for i in items:
##             self.sellist.insert(END,self.spelllist.get(i))
##         self.spelllist.select_clear(0,END)
    
##     def deselect(self):
##         items = self.sellist.curselection()
##         try:
##             items = map(int, items)
##         except ValueError: pass
##         list=[]
##         for i in self.sellist.get(0,END):
##             list.append(i)
##         for i in items:
##             list.remove(self.sellist.get(i))
##         self.sellist.delete(0,END)
##         for i in list:
##             self.sellist.insert(END,i)

##     def apply(self):
##         names=self.sellist.get(0,END)
##         for a in affects.values():
##             if a.name in names:
##                 self.triggers[a.spelldown]=(a.spelldown,"#activate "+a.name)
##                 self.triggers[a.spellup]=(a.spellup,"#disable "+a.name)
##                 make_button(a.name+"~"+"cast '"+a.name+"'")
##         self.parent.re_init()


class TriggersDialog(Dialog):
    # A dialog-class to edit a regexp with associated commands
    # Uses the TriggerFrame class
    def __init__(self, master, title,  client):
        self.client = client
        self.backuphash = self.client.triggers.copy()
        self.hash= client.triggers
        Dialog.__init__(self, master, title)
        
    def create_body(self, master):
        self.triggersframe = TriggerFrame(master,self.hash)
        self.triggersframe.pack()
        return self.triggersframe

    def apply(self):
        index=self.triggersframe.list.get(ACTIVE)
        commands = re.sub("(^ *)|( *$)","",string.replace(string.replace(self.triggersframe.text.get(1.0,END),"\n\n","\n"),"\n",";"))
        if commands[-1] == ";":
            commands=commands[:-1]
            
        self.hash[index]=(re.compile(self.triggersframe.list.get(ACTIVE)),commands)
       # self.hash.clear()
       # for k in self.localhash.keys():
       #     self.hash[k] = self.localhash[k]
        print self.hash
        print self.client.triggers.keys()
        
    def cancel(self):
        self.hash.clear()
        for k in self.backuphash.keys():
            self.hash[k]=self.backuphash[k]
        Dialog.cancel(self)

        


class TriggerFrame(HashFrame):
  def __init__(self, master, hashlist):
        Frame.__init__(self, master)
        self.name="Trigger"
        self.current=None
        self.hash=hashlist
        self.scrollbar = Scrollbar(self, orient=VERTICAL)
        self.list=Listbox(self,exportselection=0,yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.list.yview)
        self.list.grid(row=0,column=0,sticky=N+S)
        self.scrollbar.grid(row=0,column=1,sticky=N+S)
	self.list.bind("<Button-1>", self.list_has_changed)
        self.list.bind("<B1-Motion>",self.list_ignore)
        self.list.bind("<Double-Button-1>",self.list_ignore)
        self.list.bind("<ButtonRelease-1>",self.list_ignore)
	self.text=Text(self)
        self.text.grid(row=0,column=2)
        for k in self.hash.keys():
            self.list.insert(END,k)
        self.list.activate(0)
        self.list.select_clear(0,END)
        self.list.select_set(0)
        self.make_text()
        self.make_buttons("Trigger")
        
  def add_hash(self):
      self.make_hash()
      idx=self.list.index(ACTIVE)
      newname=askstring("New "+self.name,"Pattern of new "+self.name,parent=self)
      if self.hash.has_key(newname):
          if not askyesno("New "+self.name, "Overwrite old "+self.name+" "+newname+"?"):
              return
      self.hash[newname]=(re.compile(newname),"")
      self.list.delete(0, END)
      keys = self.hash.keys()
      keys.sort()
      i = 0
      for k in keys:
          self.list.insert(END,k)
          if (k == newname):
              idx = i
          i = i + 1
      self.list.activate(idx)
      self.list.select_clear(0,END)
      self.list.select_set(idx)
      self.text.delete(1.0, END)
      self.make_text()

  def make_hash(self):
      if (self.list.get(ACTIVE)):
          self.hash[self.list.get(ACTIVE)]=(re.compile(self.list.get(ACTIVE)),re.sub("(^ *)|( *$)","",string.replace(string.replace(self.text.get(1.0,END),"\n\n","\n"),"\n",";")))
  
  def make_text(self):
      	if self.list.get(ACTIVE):
            self.text.insert(END,re.sub('([^\\\]);',"\\1\\n",self.hash[self.list.get(ACTIVE)][1]))
         
  def list_ignore(self, event):
      return "break"


class AffectEditor(Dialog):
    def __init__(self, gui, client):
        self.client=client
        self.gui=gui
        #make a working-copy of the affects
        self.init_affects()
        Dialog.__init__(self,gui,"Edit Affects")
        
    def init_affects(self):
        affects = self.client.get_affects()
        self.affects={}
        for a in affects.values():
            if a.fading:
               self.affects[a.name]=Client.Affect(a.name,a.symbol,a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern,a.fading,a.standard)
            else:
                self.affects[a.name]=Client.Affect(a.name,a.symbol,a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern,standard=a.standard)
                
    def create_body(self,master):
        f1 = Frame(master)
        Label(f1,text="Affects:").pack(side=TOP,anchor=W)
        self.affect_list=Listbox(f1)
        self.affect_list.pack(side=LEFT,fill=BOTH,expand=YES)
        self.scrollbar = Scrollbar(f1, orient=VERTICAL)
        self.scrollbar.config(command=self.affect_list.yview)
        self.scrollbar.pack(side=LEFT,fill=Y)
        self.affect_list.configure(yscrollcommand=self.scrollbar.set)
        self.fill_list()
        f1.grid(column=0,row=0)
        affects=self.client.get_affects()
        f2 = Frame(master)
        
        self.name_entry=make_entry(f2,text="Name",row=1)
        self.symbol_entry=make_entry(f2,text="Symbol for display in status-bar",row=2)
        self.aff_re_entry=make_entry(f2,text="Name in affect-list",row=3)
        self.spellup_entry=make_entry(f2,text="Spellup message",row=4)
        self.spelldown_entry=make_entry(f2,text="Spelldown message",row=5)
        self.fading_entry=make_entry(f2,text="Fading message",row=6)
        f2.grid(column=1,row=0)
        self.affect_list.bind("<ButtonRelease-1>", self.affect_list_has_changed)

        f3 = Frame(master)
        self.changebutton=Button(f3,text="Change",command=self.change)
        self.changebutton.pack(side=LEFT)
        Button(f3,text="New",command=self.new).pack(side=LEFT)
        self.removebutton=Button(f3,text="Remove",command=self.remove)
        self.removebutton.pack(side=LEFT)
        Button(f3,text="Reset all",command=self.reset).pack(side=LEFT)
        f3.grid(column=0,row=1,columnspan=2)
        self.affect_list_has_changed()
        return f1

    def affect_list_has_changed(self,event=None):
        affects=self.affects
        selected = self.affect_list.curselection()
        self.index = -1
        #print "selected=",selected
        if len(selected)>0:
            index = int(selected[0])
            self.index = index
            a = affects[self.affect_list.get(index)]
            settext(self.name_entry,a.name)
            settext(self.symbol_entry,ansitostring(a.symbol))
            settext(self.aff_re_entry,a.aff_re_orig)
            settext(self.spellup_entry,a.spellup.pattern)
            settext(self.spelldown_entry,a.spelldown.pattern)
            if a.fading:
                settext(self.fading_entry,a.fading.pattern)
            else:
                settext(self.fading_entry,"")
        if self.index == -1:
            self.changebutton.configure(state=DISABLED)
            self.removebutton.configure(state=DISABLED)
        else:
            self.changebutton.configure(state=NORMAL)
            self.removebutton.configure(state=NORMAL)
        
            
   
    def change(self):
        name = ""
        a=self.affects[self.affect_list.get(self.index)]
        a.name=self.name_entry.get()
        name = a.name
        a.symbol=stringtoansi(self.symbol_entry.get())
        a.set_aff_re(self.aff_re_entry.get())
        a.spellup=re.compile(self.spellup_entry.get(), re.MULTILINE)
        a.spelldown=re.compile(self.spelldown_entry.get(), re.MULTILINE)
        a.standard=0
        if len(string.strip(self.fading_entry.get()))>0:
            a.fading=re.compile(self.fading_entry.get(), re.MULTILINE)
        else:
            a.fading=None
        self.fill_list()
        index=list(self.affect_list.get(0,END)).index(name)
        self.affect_list.select_set(index)
        self.affect_list.see(index)
        self.affect_list_has_changed()

    def reset(self):
        if askyesno("Reset affects","All custom affects will be deleted. Continue?"):
            self.client.resetaffects()
            self.init_affects()
            self.fill_list()
            self.affect_list_has_changed()
        
    def apply(self):
        self.client.set_affects(self.affects)
        
    def new(self):
        newname=askstring("New Affect","Name of new affect")
        newaff = None
        if self.affects.has_key(newname):
                if askyesno("New Affect", "Overwrite old Affect "+newname+"?"):
                    newaff=a
                    pass
                else:
                    return
        if not newaff:
            newaff=Client.Affect(newname,"",newname,"","")
            self.affects[newname]=newaff
        else:
            newaff.symbol=re.compile("", re.MULTILINE)
            newaff.set_aff_re=("")
            newaff.spellup=re.compile("", re.MULTILINE)
            newaff.spelldown=re.compile("", re.MULTILINE)
            newaff.fading=None
        newaff.standard=0
        self.fill_list()
        index=list(self.affect_list.get(0,END)).index(newname)
        self.affect_list.select_set(index)
        self.affect_list.see(index)
        self.affect_list_has_changed()

    def remove(self):
        if self.affects[self.affect_list.get(self.index)].standard:
            showinfo("Removing a standard affect","Standard affects cannot permanently be removed.")
        del self.affects[self.affect_list.get(self.index)]
        self.fill_list()
        self.affect_list_has_changed()
        settext(self.name_entry,"")
        settext(self.symbol_entry,"")
        settext(self.aff_re_entry,"")
        settext(self.spellup_entry,"")
        settext(self.spelldown_entry,"")
        settext(self.fading_entry,"")
        
    def fill_list(self):
        self.affect_list.delete(0,END)
        for aff in self.affects.values():
            self.affect_list.insert(END, aff.name)
   
        
           
                               
                       
           
               
               
           
        

class SpellButtonEditor(Toplevel):
    def __init__(self, client, gui):
        Toplevel.__init__(self)
        self.title("Edit Buttons")
        self.client = client
        self.gui = gui
        
        affects = client.get_affects()
        self.index = -1

        #create a backup of the spellbuttons
        self.backupbuttons = []
        for b in self.gui.affect_buttons:
            self.backupbuttons.append([b.name,b.command])
        
        f = Frame (self)
        f2 = Frame(f)
        f3 = Frame(f2)
        Label(f3,text="Affects:").pack(side=TOP,anchor=W)
        list = Listbox(f3)
        list.pack(side=LEFT,fill=BOTH,expand=YES)
        self.affect_list = list
        self.scrollbar = Scrollbar(f3, orient=VERTICAL)
        self.scrollbar.config(command=self.affect_list.yview)
        self.scrollbar.pack(side=LEFT,fill=Y)
        list.configure(yscrollcommand=self.scrollbar.set)
        f3.pack(side=TOP,expand=YES,fill=BOTH)
        f2.pack(side=LEFT, fill=BOTH,expand=YES)
        v = affects.keys()
        v.sort()
        for aff in v: #affects.values():
            list.insert(END, aff)

        f2 = Frame(f)
        f3 = Frame(f2)
        Label(f3,text="Buttons:").pack(side=TOP,anchor=W)
        list = Listbox(f3)
        list.pack(side=LEFT,fill=BOTH,expand=YES)
        self.button_list = list
        self.scrollbar = Scrollbar(f3, orient=VERTICAL)
        self.scrollbar.config(command=self.button_list.yview)
        self.scrollbar.pack(side=LEFT,fill=Y)
        list.configure(yscrollcommand=self.scrollbar.set)
        f3.pack(side=TOP,expand=YES,fill=BOTH)
        f2.pack(side=LEFT, fill=BOTH,expand=YES)

        self.fill_buttons()
        
        #f.grid(row=0,column=0,sticky=W)
        f.pack(side=TOP, fill=BOTH,expand=YES)

        f = Frame(self)
        Label(f,text="Name:       ").pack(side=LEFT)
        entry = Entry(f)
        self.name_entry = entry
        entry.pack(side=LEFT, fill=X,expand = YES)
        f.pack(side=TOP, fill=X)

        f = Frame(self)
        Label(f,text="Command: ").pack(side=LEFT)
        entry = Entry(f)
        self.command_entry = entry
        entry.pack(side=LEFT, fill=X,expand = YES)
        f.pack(side=TOP, fill=X)

        
        f = Frame(self)
        b = Button(f,text="ok", command=self.do_ok)
        b.pack(side=LEFT)


        b = Button(f,text="add", command=self.add)
        b.pack(side=LEFT)

        b = Button(f,text="change", command=self.do_change)
        b.pack(side=LEFT)
        self.change_button = b
        
        b = Button(f,text="remove", command=self.do_remove)
        b.pack(side=LEFT)
        self.remove_button = b
        
        b = Button(f,text="cancel", command=self.do_cancel)
        self.bind("<Escape>", self.do_cancel)
        
        b.pack(side=LEFT)
        f.pack(side=TOP, fill=X)
        self.cancel_button = b

        self.remove_button.configure(state=DISABLED)
        self.change_button.configure(state=DISABLED)
        self.affect_list.bind("<ButtonRelease-1>", self.affect_list_has_changed)
        self.button_list.bind("<ButtonRelease-1>", self.button_list_has_changed)
        
    def affect_list_has_changed(self,event):
        selected = self.affect_list.curselection()
        self.index = -1
        #print "selected=",selected
        if len(selected)>0:
            index = int(selected[0])
            self.index = index
            self.name_entry.delete(0,END)
            self.name_entry.insert(END, self.affect_list.get(index))
            self.command_entry.delete(0,END)
            self.command_entry.insert(END, 'cast "'+self.affect_list.get(index)+'"')

            self.remove_button.configure(state=DISABLED)
            self.change_button.configure(state=DISABLED)

    def button_list_has_changed(self,event):
        selected = self.button_list.curselection()
        self.index = -1
        #print "selected=",selected
        if len(selected)>0:
            index = int(selected[0])
            self.index = index
            self.name_entry.delete(0,END)
            name = self.button_list.get(index)
            self.name_entry.insert(END, name)
            self.command_entry.delete(0,END)
            for b in self.gui.affect_buttons:
                if b.name == name:
                    self.command_entry.insert(END, b.command)
                    break

            self.remove_button.configure(state=NORMAL)
            self.change_button.configure(state=NORMAL)                        

    def fill_buttons(self):
        self.button_list.delete(0,END)
        for b in self.gui.affect_buttons:
            self.button_list.insert(END, b.name)
        
    def add(self):
        spell = self.name_entry.get()
        txt = self.command_entry.get()
        if len(spell)>0 and len(txt)>0:
            self.gui.create_affect_button(spell, txt)
            self.fill_buttons()            
            
    def do_remove(self):
        selected = self.button_list.curselection()
        self.index = -1
        if len(selected)>0:
            index = int(selected[0])
            name = self.button_list.get(index)
            #for b in self.gui.affect_buttons:
            #    if b.name == name:
            #        b.destroy()
            #        self.gui.affect_buttons.remove(b)
            #call the gui-method to remove the spellframe if necessary
            self.gui.remove_affect_button(name)
            self.remove_button.configure(state=DISABLED)
            self.change_button.configure(state=DISABLED)

            self.fill_buttons()
            #break

    def do_cancel(self,event=None):
        #remove all spellbuttons
        self.gui.clear_affect_buttons()
        #and create them anew from the backup
        for b in self.backupbuttons:
            self.gui.create_affect_button(b[0],b[1])
        self.gui.focus_set()
        self.destroy()
        
    def do_ok(self):
        self.gui.focus_set()
        self.destroy()
        
    def do_change(self):
        selected = self.button_list.curselection()
        self.index = -1
        if len(selected)>0:
            index = int(selected[0])
            name = self.button_list.get(index)
            new_name = self.name_entry.get()
            new_command=self.command_entry.get()
            for b in self.gui.affect_buttons:
                if b.name == name:
                    b.configure(text=new_name)
                    b.name = new_name
                    b.command = new_command
                    
                    self.remove_button.configure(state=DISABLED)
                    self.change_button.configure(state=DISABLED)

                    self.fill_buttons()
                    break
                
class ColorButton(Button):
    def __init__(self,parent):
        Button.__init__(self,parent,command=self.do_color)
        
    def do_color(self):
        chosen = tkColorChooser.askcolor(initialcolor=self.cget("background"),title="Select color")
        if chosen != (None, None):
# askcolor returns a tupel, with the second item being the tk colorstring
            self.config(background=chosen[1])
    
class ColorEditor(Dialog):
    def __init__(self,parent,title,client):
        self.client=client
        Dialog.__init__(self,parent,title)
        
    def create_body(self,master):
        #create local copies of the color-definitions
        self.button={}
        #"big" frame so we can use f.pack()
        f1 = Frame(master,relief=RIDGE,borderwidth=1) 
        f1.grid(column=0,row=0,columnspan=1,sticky=N+S+W+E)
        f = Frame(f1,relief=RIDGE,borderwidth=1)
        
        #create the buttons for the normal colors
        j=0
        for i in Term.rgb.keys():
            self.button[i]=ColorButton(f)
            self.button[i].configure(background=Term.rgb[i])
            self.button[i].grid(column=0,row=j)
            Label(f,text=i).grid(column=1,row=j)
            j = j+ 1
        j=0
        
        #create the buttons for the bright colors
        f.pack(side=LEFT)
        f = Frame(f1,relief=RIDGE,borderwidth=1)
        for i in Term.brgb.keys():
            self.button["bright "+i]=ColorButton(f)
            self.button["bright "+i].configure(background=Term.brgb[i])
            self.button["bright "+i].grid(column=0,row=j)
            Label(f,text="bright "+i).grid(column=1,row=j)
            j = j+ 1
        f.pack(side=LEFT)
        f = Frame(f1,relief=RIDGE,borderwidth=1)
        self.foreground=self.client.gui.term.text.tag_cget("normal","foreground")
        self.background=self.client.gui.term.text.tag_cget("normal","background")
        self.affect_button_active_fg = self.client.gui.affect_button_active_fg
        self.affect_button_active_bg = self.client.gui.affect_button_active_bg
        self.affect_button_fading_fg=self.client.gui.affect_button_fading_fg
        self.affect_button_fading_bg=self.client.gui.affect_button_fading_bg
        self.affect_button_inactive_fg=self.client.gui.affect_button_inactive_fg
        self.affect_button_inactive_bg=self.client.gui.affect_button_inactive_bg
        self.fgbutton=ColorButton(f)
        self.fgbutton.configure(background=self.foreground)
        self.fgbutton.grid(column=0,row=0)
        Label(f,text="Foreground").grid(column=1,row=0)
        self.bgbutton=ColorButton(f)
        self.bgbutton.configure(background=self.background)
        self.bgbutton.grid(column=0,row=1)
        Label(f,text="Background").grid(column=1,row=1)
        
        f.pack(side=LEFT)
        f1 = Frame(master,relief=RIDGE,borderwidth=1) 
        f1.grid(column=0,row=1,sticky=N+S+W+E)
        Label(f1,text="Spellbutton colors").grid(column=0,row=0,columnspan=3,sticky=W+E)
        Label(f1,text="Foreground").grid(column=1,row=1,sticky=W+E)
        Label(f1,text="Background").grid(column=2,row=1,sticky=W+E)
        Label(f1,text="Active").grid(column=0,row=2,sticky=W+E)
        Label(f1,text="Faded").grid(column=0,row=3,sticky=W+E)
        Label(f1,text="Inactive").grid(column=0,row=4,sticky=W+E)
        self.affect_fa=ColorButton(f1)
        self.affect_fa.configure(background=self.affect_button_active_fg)
        self.affect_fa.grid(column=1,row=2,sticky=W+E)
        self.affect_ba=ColorButton(f1)
        self.affect_ba.configure(background=self.affect_button_active_bg)
        self.affect_ba.grid(column=2,row=2,sticky=W+E)
        self.affect_ff=ColorButton(f1)
        self.affect_ff.configure(background=self.affect_button_fading_fg)
        self.affect_ff.grid(column=1,row=3,sticky=W+E)
        self.affect_bf=ColorButton(f1)
        self.affect_bf.configure(background=self.affect_button_fading_bg)
        self.affect_bf.grid(column=2,row=3,sticky=W+E)
        self.affect_fi=ColorButton(f1)
        self.affect_fi.configure(background=self.affect_button_inactive_fg)
        self.affect_fi.grid(column=1,row=4,sticky=W+E)
        self.affect_bi=ColorButton(f1)
        self.affect_bi.configure(background=self.affect_button_inactive_bg)
        self.affect_bi.grid(column=2,row=4,sticky=W+E)
         
        
        return f1
    
    def apply(self):
        for k in Term.brgb.keys():
            Term.brgb[k]=self.button["bright "+k].cget("background")
        for k in Term.rgb.keys():
            Term.rgb[k]=self.button[k].cget("background")
        self.client.gui.affect_button_active_fg = self.affect_fa.cget("background")
        self.client.gui.affect_button_active_bg = self.affect_ba.cget("background")
        self.client.gui.affect_button_fading_fg = self.affect_ff.cget("background")
        self.client.gui.affect_button_fading_bg = self.affect_bf.cget("background")
        self.client.gui.affect_button_inactive_fg = self.affect_fi.cget("background")
        self.client.gui.affect_button_inactive_bg = self.affect_bi.cget("background")
        self.client.gui.color_buttons()
        bg=self.bgbutton.cget("background")
        text = self.client.gui.term.text
        for c in Term.colors:
            text.tag_configure(c,foreground=Term.rgb[c],background=bg)
            text.tag_configure("b"+c,foreground=Term.brgb[c],background=bg)
            text.tag_configure("v"+c,background=Term.rgb[c],foreground=bg)
            text.tag_configure("vb"+c,background=Term.brgb[c],foreground=bg)
        self.client.gui.term.text.tag_configure("normal",foreground=self.fgbutton.cget("background"),background=bg)
        self.client.gui.term.text.configure(background=bg)
        self.client.gui.status.configure(background=bg)
        self.client.gui.aff_win.text.configure(background=bg)


    
class PreferenceDialog(Dialog):
    def __init__(self, parent,title,client):
        self.client=client
        Dialog.__init__(self,parent,title)
        
    def create_body(self, master):
        self.systemopts=Frame(master)
        self.triggers_enabled=IntVar()
        self.triggers_enabled.set(self.client.triggers_on)
        Checkbutton(self.systemopts,text="Enable Triggers",variable=self.triggers_enabled).grid(sticky=N+W,column=0,row=0,columnspan=2)
        self.echovar = IntVar()
        self.echovar.set(self.client.echo_commands)
        Checkbutton(self.systemopts,text="Local echo",variable=self.echovar).grid(sticky=N+W,column=0,row=2,columnspan=2)

        self.textopts=Frame(self.systemopts)
        self.textopts.grid(column=0,sticky=W+E)
        self.logfile=StringVar()
        if (self.client.log):
            self.logfile.set(self.client.log.name)
        else:
            self.logfile.set("")
        Label(self.textopts,text="Logfile:").grid(sticky=W)
        Entry(self.textopts,textvariable=self.logfile).grid(sticky=W+E)
        self.prefix=StringVar()

        if self.client.prefix:
            self.prefix.set(self.client.prefix)
        Label(self.textopts,text="Prefix:").grid(sticky=W)
        Entry(self.textopts,textvariable=self.prefix).grid(sticky=E+W)

        
        self.recallvar = IntVar()
        self.recallvar.set(self.client.recall_size)
        Label(self.textopts,text="Recall size").grid(sticky=W)
        Entry(self.textopts,textvariable=self.recallvar).grid(sticky=E+W)
        
        self.scrollsizevar = IntVar()
        self.scrollsizevar.set(self.client.get_ioption("scrollsize"))
        Label(self.textopts,text="Ammount to scroll on PgUp/Down").grid(sticky=W)
        Entry(self.textopts,textvariable=self.scrollsizevar).grid(sticky=E+W)
              
        self.textopts.grid(column=1,row=1)
        self.systemopts.pack()
        return self.systemopts

    def apply(self):
        self.client.triggers_enabled=self.triggers_enabled.get()
        self.client.echo_commands = self.echovar.get()
        Client.do_log(self.client,self.logfile.get())
        Client.do_prefix(self.client,self.prefix.get())
        Client.set_recall_size(self.client,self.recallvar.get())
        self.client.set_option("scrollsize",self.scrollsizevar.get())
        #menu_orient=self.MOVar.get()
        #custom_orient=self.BOVar.get()
        
        #for b in buttons.hash.keys():
        #    buttons.hash[b].button.pack(side=custom_orient)

class ColorCodeButton(Button):

    def __init__(self,parent,gui, code):
        self.gui = gui
        self.code = code

        Button.__init__(self,parent,text=code,command=self.do_color)

    def do_color(self):
        entry = self.gui.entry
        #print entry.index(INSERT)
        if entry.selection_present():
            #print entry.selection_get()
            #print entry.index(SEL_FIRST)
            #print entry.index(SEL_LAST)
            
            self.gui.entry.insert(SEL_FIRST,self.code)
            self.gui.entry.insert(SEL_LAST,"#n")

        else:
            entry.insert(INSERT,self.code)
        entry.focus_set()


class ColorTool(Toplevel):
    def __init__(self,  gui):
        Toplevel.__init__(self)
        self.title("Color Code")
        self.gui = gui
        self.transient(gui)

        self.code = { "black":  "#l",
        "red":    "#r",
        "green":  "#g",
        "yellow": "#y",
        "blue":   "#b",
        "magenta":"#m",
        "cyan":   "#c",
        "white":  "#w" }
        self.code_bright = { "black":  "#L",
        "red":    "#R",
        "green":  "#G",
        "yellow": "#Y",
        "blue":   "#B",
        "magenta":"#M",
        "cyan":   "#C",
        "white":  "#W" }
        
        self.button={}

        #"big" frame so we can use f.pack()
        f1 = Frame(self,relief=RIDGE,borderwidth=1) 
        f1.grid(column=0,row=0,columnspan=1)
        f = Frame(f1,relief=RIDGE,borderwidth=1)
        
        #create the buttons for the normal colors
        j=0
        for i in Term.rgb.keys():
            self.button[i]=ColorCodeButton(f,gui,self.code[i])
            self.button[i].configure(background=Term.rgb[i])
            self.button[i].grid(column=0,row=j)
            #Label(f,text=i).grid(column=1,row=j)
            j = j+ 1
        j=0
        
        #create the buttons for the bright colors
        f.pack(side=LEFT)
        f = Frame(f1,relief=RIDGE,borderwidth=1)
        for i in Term.brgb.keys():
            self.button["bright "+i]=ColorCodeButton(f,gui,self.code_bright[i])
            self.button["bright "+i].configure(background=Term.brgb[i])
            self.button["bright "+i].grid(column=0,row=j)
            #Label(f,text="bright "+i).grid(column=1,row=j)
            j = j+ 1
        f.pack(side=LEFT)

class CharEditor(Dialog):
    def __init__(self,parent,gui,title):
        self.gui=gui
        Dialog.__init__(self,parent,title)

    def create_body(self,master):
        box = Frame(master)
        
